#include <WiFi.h>
#include <ESP32Servo.h>
#include <Keypad.h>
#include <HTTPClient.h>


// WiFi credentials
#include "wifi_codes.h"

// Define locker ID
const char* locker_id = "a";  // This is locker "a"

// Server URLs
const char* statusServerURL = "http://10.29.226.47:8080/update_locker";  // For locker status updates
// const char* statusServerURL = "https://buybye-backend.onrender.com/update_locker";  // For locker status updates - deployed
const char* openCodeURL = "http://10.29.226.47:8080/get_open_code?locker_id=a";

// Servo and LED Pins
int servoPin = 44;
int ledPin = 7;
int green = 8;
int red = 9;

Servo myServo;

// Keypad configuration
const byte ROWS = 4; // 4 rows
const byte COLS = 3; // 3 columns
char keys[ROWS][COLS] = {
  {'1','2','3'},
  {'4','5','6'},
  {'7','8','9'},
  {'*','0','#'}
};
byte rowPins[ROWS] = {1, 2, 3, 4};
byte colPins[COLS] = {5, 6, 43};
Keypad keypad = Keypad(makeKeymap(keys), rowPins, colPins, ROWS, COLS);

// Define access codes
String open_code = "1234";
String close_code = "*";
String inputCode = "";



void setup() {
  Serial.begin(115200);
  WiFi.begin(ssid, password);
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }
  Serial.println("\nWiFi connected.");
  Serial.print("IP Address: ");
  Serial.println(WiFi.localIP());


  fetchOpenCodeFromServer();

  // Servo initialization
  myServo.attach(servoPin);
  myServo.write(90); // Locked position

  // LED initialization
  pinMode(ledPin, OUTPUT);
  pinMode(green, OUTPUT);
  pinMode(red, OUTPUT);
  digitalWrite(ledPin, LOW);
  digitalWrite(green, LOW);
  digitalWrite(red, LOW);

}


void loop() {
  char key = keypad.getKey();  // Get keypad input
  if (key) {
    if (key == '#') {  // '#' triggers code submission
      if (inputCode == open_code) {
        openLocker();
      } else {
        Serial.println("Invalid code.");
        indicateError();
      }
      inputCode = "";  // Reset input after submission
    } else if (key == '*') {  // '*' closes locker
        closeLocker();
    } else {
      inputCode += key;  // Append key to input
      Serial.println("Entered: " + inputCode);
    }
  }

}

void openLocker() {
  Serial.println("Opening locker...");
  digitalWrite(green, HIGH);  // Turn on green LED
  myServo.write(0);         // Open servo
  sendLockerStatus(1);        // Send status: open
  delay(2000);                
  digitalWrite(green, LOW);   // Turn off green LED
}

void closeLocker() {
  Serial.println("Closing locker...");
  // digitalWrite(ledPin, HIGH);  // Turn on locker light
  myServo.write(90);            // Lock servo
  sendLockerStatus(0);        // Send status: closed

  delay(2000);                 // Keep light on for 2 seconds
  // digitalWrite(ledPin, LOW);   // Turn off locker light
  fetchOpenCodeFromServer();
}

void indicateError() {
  digitalWrite(red, HIGH);  // Turn on red LED
  delay(1000);              // Keep red LED on for 1 second
  digitalWrite(red, LOW);   // Turn off red LED
}


void sendLockerStatus(int status) {
  if (WiFi.status() == WL_CONNECTED) {
    const int maxRetries = 5;      // Maximum number of retry attempts
    const int retryDelay = 2000;   // Delay between retries in milliseconds
    int attempt = 0;
    bool success = false;

    // String payload = "{\"locker_id\": " + String(locker_id) + ", \"status\": " + String(status) + "}";
    String payload = "{\"locker_id\": \"" + String(locker_id) + "\", \"status\": " + String(status) + "}";


    // Debug: Print payload being sent
    Serial.println("Payload being sent:");
    Serial.println(payload);

    while (attempt < maxRetries && !success) {
      Serial.printf("Attempt %d to send locker status...\n", attempt + 1);
      Serial.printf("WiFi Signal Strength (RSSI): %d dBm\n", WiFi.RSSI());
      Serial.printf("Free heap memory: %d bytes\n", ESP.getFreeHeap());

      HTTPClient http;  // Create a new HTTPClient for each attempt

      http.begin(statusServerURL);  // Initialize the connection
      http.addHeader("Content-Type", "application/json");
      http.addHeader("Content-Length", String(payload.length()));  // Add Content-Length header


      // Debug: Print payload
      Serial.println("Sending payload:");
      Serial.println(payload);

      int httpResponseCode = http.POST(payload);  // Send the HTTP POST request

      Serial.print("HTTP Response Code: ");
      Serial.println(httpResponseCode);

      if (httpResponseCode > 0) {
        // Success: Print the server's response
        String response = http.getString();
        Serial.println("Response:");
        Serial.println(response);
        success = true;  // Mark success
      } else {
        // Failed: Print error and retry
        Serial.printf("HTTP POST failed, code: %d. Retrying...\n", httpResponseCode);
        delay(retryDelay);  // Wait before retrying
        attempt++;
      }

      http.end();  // Close the connection after each attempt
    }

    if (!success) {
      // After maximum retries
      Serial.println("Failed to send locker status after maximum retries.");
    }
  } else {
    // WiFi is not connected
    Serial.println("WiFi not connected, attempting reconnection...");
    WiFi.begin(ssid, password);  // Try reconnecting to WiFi
    delay(1000);                 // Allow time for reconnection
  }
}

void fetchOpenCodeFromServer() {
  if (WiFi.status() == WL_CONNECTED) {
    HTTPClient http;
    http.begin(openCodeURL);

    int httpResponseCode = http.GET();
    if (httpResponseCode > 0) {
      String payload = http.getString();
      Serial.println("Open code response payload:");
      Serial.println(payload);

      int codeStart = payload.indexOf("\"open_code\":\"");
      if (codeStart != -1) {
        codeStart += strlen("\"open_code\":\"");
        int codeEnd = payload.indexOf("\"", codeStart);
        if (codeEnd != -1) {
          String newCode = payload.substring(codeStart, codeEnd);
          Serial.print("Updating open_code to: ");
          Serial.println(newCode);
          open_code = newCode; // Update the global open_code variable
        } else {
          Serial.println("Error: Could not find end quote for open_code.");
        }
      } else {
        Serial.println("Error: open_code not found in the response.");
      }
    } else {
      Serial.print("GET request failed, code: ");
      Serial.println(httpResponseCode);
    }

    http.end();
  } else {
    Serial.println("WiFi not connected, cannot fetch open_code.");
  }
}

